# distutils: language = c++
# cython: language_level = 3
cimport numpy as np
import numpy as np

np.import_array()

from runtime.runtime cimport BatchMailBox, NullResult, Scheduler, WaitResult
from libc.math cimport log2, fmax, fmin, fabs, ceil
from libc.stdio cimport printf
from libc.stdlib cimport malloc, free
from openmp cimport omp_get_max_threads
from cython.operator cimport dereference as deref
from cython.parallel import prange

## Types declaration
ctypedef int I_t
ctypedef double D_t

cdef lock Scheduler scheduler
cdef D_t INF = 1e38

cdef struct NodeData_t:
    I_t idx_start
    I_t idx_end
    I_t is_leaf

# NOTE: The following extern definition is used to interface
# std::nth_element, a robust partitioning algorithm, in Cython
cdef extern from *:
    """
    #include <algorithm>
    #include <iostream>
    template<class D, class I>
    class IndexComparator {
    private:
        const D *data;
        I split_dim, n_features;
    public:
        IndexComparator(const D *data, const I &split_dim, const I &n_features):
            data(data), split_dim(split_dim), n_features(n_features) {}
        bool operator()(const I &a, const I &b) const {
            D a_value = data[a * n_features + split_dim];
            D b_value = data[b * n_features + split_dim];
            return a_value == b_value ? a < b : a_value < b_value;
        }
    };
    template<class D, class I>
    void partition_node_indices(
        const D *data,
        I *node_indices,
        I start,
        I mid,
        I end,
        const I &split_dim,
        const I &n_features) {
        IndexComparator<D, I> index_comparator(data, split_dim, n_features);

        std::nth_element(
            node_indices + start,
            node_indices + mid,
            node_indices + end,
            index_comparator
        );

    }
    """
    void partition_node_indices[D, I](
                D *data_ptr,
                I *node_indices,
                I start,
                I mid,
                I end,
                I split_dim,
                I n_features
    ) nogil except +

cdef I_t find_node_split_dim(D_t* data,
                             I_t* node_indices,
                             I_t n_features,
                             I_t n_points) nogil except -1:
    """Find the dimension with the largest spread.

    Parameters
    ----------
    data : double pointer
        Pointer to a 2D array of the training data, of shape (n_samples, n_features).
        n_samples must be greater than any of the values in node_indices.
    node_indices : int pointer
        Pointer to a 1D array of length n_points.  This lists the indices of
        each of the points within the current node.

    Returns
    -------
    j_max : int
        The index of the feature (dimension) within the node that has the
        largest spread.

    Notes
    -----
    In numpy, this operation is equivalent to

    def find_node_split_dim(data, node_indices):
        return np.argmax(data[node_indices].max(0) - data[node_indices].min(0))

    The cython version is much more efficient in both computation and memory.
    """
    cdef D_t min_val, max_val, val, spread, max_spread
    cdef I_t i, j, j_max

    j_max = 0
    max_spread = 0

    for j in range(n_features):
        max_val = data[node_indices[0] * n_features + j]
        min_val = max_val
        for i in range(1, n_points):
            val = data[node_indices[i] * n_features + j]
            max_val = fmax(max_val, val)
            min_val = fmin(min_val, val)
        spread = max_val - min_val
        if spread > max_spread:
            max_spread = spread
            j_max = j
    return j_max


cdef cypclass Counter activable:
    """ A simple Counter.

    This can be useful for synchronisation for the caller after
    triggering the actors logic as it wait for the value of
    the Counter to reach a given one before moving on.
    """

    I_t _n

    __init__(self):
        self._n = 0
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)

    void add(self, I_t value):
        self._n += value

    void reset(self):
        self._n = 0

    I_t value(self):
        return self._n


cdef cypclass Container activable:
    """ A simple wrapper of an array.

    The wrapped array is passed by the initial caller which then
    trigger the actors logic modifying it.

    The initial caller can wait for specific number of update
    before proceeding.
    """

    I_t * _array
    I_t _size
    I_t _n_updates


    __init__(self, I_t *array, I_t size):
        self._array = array
        self._size = size
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)
        self._n_updates = 0

    void update(self, I_t idx, I_t value):
        self._n_updates += 1
        self._array[idx] = value

    I_t n_updates(self):
        return self._n_updates


cdef inline D_t sqeuclidean_dist(D_t* x1, D_t* x2, I_t k) nogil:
    cdef:
        I_t i = 0
        D_t distance = 0

    for i in range(k):
        distance += (x1[i] - x2[i]) ** 2

    return distance


cdef inline void dual_swap(D_t* darr, I_t* iarr, I_t i1, I_t i2) nogil:
    """swap the values at inex i1 and i2 of both darr and iarr"""
    cdef D_t dtmp = darr[i1]
    darr[i1] = darr[i2]
    darr[i2] = dtmp

    cdef I_t itmp = iarr[i1]
    iarr[i1] = iarr[i2]
    iarr[i2] = itmp


cdef void _simultaneous_sort(
    D_t* dist,
    I_t* idx,
    I_t size
) nogil:
    """
    Perform a recursive quicksort on the dist array, simultaneously
    performing the same swaps on the idx array.  The equivalent in
    numpy (though quite a bit slower) is

    def simultaneous_sort(dist, idx):
        i = np.argsort(dist)
        return dist[i], idx[i]
    """
    cdef I_t pivot_idx, i, store_idx
    cdef D_t pivot_val

    # in the small-array case, do things efficiently
    if size <= 1:
        pass
    elif size == 2:
        if dist[0] > dist[1]:
            dual_swap(dist, idx, 0, 1)
    elif size == 3:
        if dist[0] > dist[1]:
            dual_swap(dist, idx, 0, 1)
        if dist[1] > dist[2]:
            dual_swap(dist, idx, 1, 2)
            if dist[0] > dist[1]:
                dual_swap(dist, idx, 0, 1)
    else:
        # Determine the pivot using the median-of-three rule.
        # The smallest of the three is moved to the beginning of the array,
        # the middle (the pivot value) is moved to the end, and the largest
        # is moved to the pivot index.
        pivot_idx = size // 2
        if dist[0] > dist[size - 1]:
            dual_swap(dist, idx, 0, size - 1)
        if dist[size - 1] > dist[pivot_idx]:
            dual_swap(dist, idx, size - 1, pivot_idx)
            if dist[0] > dist[size - 1]:
                dual_swap(dist, idx, 0, size - 1)
        pivot_val = dist[size - 1]

        # partition indices about pivot.  At the end of this operation,
        # pivot_idx will contain the pivot value, everything to the left
        # will be smaller, and everything to the right will be larger.
        store_idx = 0
        for i in range(size - 1):
            if dist[i] < pivot_val:
                dual_swap(dist, idx, i, store_idx)
                store_idx += 1
        dual_swap(dist, idx, store_idx, size - 1)
        pivot_idx = store_idx

        # recursively sort each side of the pivot
        if pivot_idx > 1:
            _simultaneous_sort(dist, idx, pivot_idx)
        if pivot_idx + 2 < size:
            _simultaneous_sort(dist + pivot_idx + 1,
                                    idx + pivot_idx + 1,
                                    size - pivot_idx - 1)


cdef cypclass NeighborsHeaps:
    """A max-heap structure to keep track of distances/indices of neighbors

    This implements an efficient pre-allocated set of fixed-size heaps
    for chasing neighbors, holding both an index and a distance.
    When any row of the heap is full, adding an additional point will push
    the furthest point off the heap.

    Taken and adapted from:
    https://github.com/scikit-learn/scikit-learn/blob/e4bb9fa86b0df873ad750b6d59090843d9d23d50/sklearn/neighbors/_binary_tree.pxi#L513

    Parameters
    ----------
    n_pts : int
        the number of heaps to use
    n_nbrs : int
        the size of each heap.
    """
    I_t *_indices
    D_t *_distances

    I_t _n_pts
    I_t _n_nbrs
    I_t _n_pushes
    bint _sorted

    __init__(self,
        I_t * indices,
        D_t * distances,
        I_t n_pts,
        I_t n_nbrs,
    ):
        self._n_pts = n_pts
        self._n_nbrs = n_nbrs
        self._distances = distances
        self._indices = indices
        self._n_pushes = 0
        self._sorted = False
        cdef I_t i

        # We can't use memset here
        for i in range(n_pts * n_nbrs):
            self._distances[i] = INF

    void push(self, I_t row, D_t val, I_t i_val):
        """push (val, i_val) into the given row"""
        cdef:
            I_t i, left_child_idx, right_child_idx, swap_idx

            # Getting the heap to use
            I_t *indices = self._indices + row * self._n_nbrs
            D_t *distances = self._distances + row * self._n_nbrs

        self._n_pushes += 1

        # check if val should be in heap
        if val > distances[0]:
            return

        # insert val at position zero
        distances[0] = val
        indices[0] = i_val

        # descend the heap, swapping values until the max heap criterion is met
        i = 0
        while True:
            left_child_idx = 2 * i + 1
            right_child_idx = left_child_idx + 1

            if left_child_idx >= self._n_nbrs:
                break
            elif right_child_idx >= self._n_nbrs:
                if distances[left_child_idx] > val:
                    swap_idx = left_child_idx
                else:
                    break
            elif distances[left_child_idx] >= distances[right_child_idx]:
                if val < distances[left_child_idx]:
                    swap_idx = left_child_idx
                else:
                    break
            else:
                if val < distances[right_child_idx]:
                    swap_idx = right_child_idx
                else:
                    break

            distances[i] = distances[swap_idx]
            indices[i] = indices[swap_idx]

            i = swap_idx

        distances[i] = val
        indices[i] = i_val

    void sort(self):
        # NOTE: Ideally we could sort results in parallel, but
        # OpenMP threadpool and this runtime's aren't working
        # nicely together (using prange here would create OpenMP
        # threads within workers, causing unexpected behavior.)
        cdef I_t row

        for row in range(self._n_pts):
            # We use a function here to be able to recurse
            _simultaneous_sort(
                self._distances + row * self._n_nbrs,
                self._indices + row * self._n_nbrs,
                self._n_nbrs,
            )

        self._sorted = True

    int n_pushes(self):
        return self._n_pushes

    int is_sorted(self):
        # TODO: is there a support for returning bool via
        # promises? As of now returning a int for making
        # use of getIntResult
        return 1 if self._sorted else 0

    D_t largest(self, I_t index):
        return self._distances[index * self._n_nbrs]


cdef cypclass Node activable:
    """A KDTree Node

    Node delegates tasks to their children Nodes.

    Some Nodes are set as Leaves when they are associated
    to ``leaf_size`` or less points.

    Leafs are terminal Nodes and do not have children.
    """
    NodeData_t *_node_data_ptr
    D_t * _node_bounds_ptr

    # Reference to the head of the allocated arrays data gets not modified via _data_ptr
    I_t node_index

    active Node _left
    active Node _right

    I_t _node_bounds_ptr_offset

    __init__(self, NodeData_t * node_data_ptr, D_t * node_bounds_ptr, I_t node_bounds_ptr_offset):
        # Needed by for Cython+ actors
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)

        self._node_data_ptr = node_data_ptr
        self._node_bounds_ptr = node_bounds_ptr

        # Offset between min and max
        self._node_bounds_ptr_offset = node_bounds_ptr_offset

    # We use this to allow using actors for initialisation
    # because __init__ can't be reified.
    void build_node(
        self,
        I_t node_index,
        D_t * data_ptr,
        I_t * indices_ptr,
        I_t leaf_size,
        I_t n_features,
        I_t dim,
        I_t idx_start,
        I_t idx_end,
        active Counter counter,
    ):
        cdef NodeData_t * node_data = self._node_data_ptr + node_index
        deref(node_data).idx_start = idx_start
        deref(node_data).idx_end = idx_end
        deref(node_data).is_leaf = False

        cdef D_t * lower_bounds = (
            self._node_bounds_ptr + node_index * n_features
        )
        cdef D_t * upper_bounds = (
            lower_bounds + self._node_bounds_ptr_offset
        )

        cdef D_t * data_row

        # Determine Node bounds
        for j in range(n_features):
            lower_bounds[j] = INF
            upper_bounds[j] = -INF

        # Compute the actual data range.  At build time, this is slightly
        # slower than using the previously-computed bounds of the parent node,
        # but leads to more compact trees and thus faster queries.
        for i in range(idx_start, idx_end):
            data_row = data_ptr + indices_ptr[i] * n_features
            for j in range(n_features):
                lower_bounds[j] = fmin(lower_bounds[j], data_row[j])
                upper_bounds[j] = fmax(upper_bounds[j], data_row[j])

        # Choose the dimension with maximum spread at each recursion instead.
        cdef I_t next_dim = find_node_split_dim(
            data=data_ptr,
            node_indices=indices_ptr + idx_start,
            n_features=n_features,
            n_points=idx_end - idx_start,
        )
        cdef I_t mid = (idx_start + idx_end) // 2

        if idx_end - idx_start <= leaf_size:
            deref(node_data).is_leaf = True
            # Adding to the global counter the number
            # of samples the leaf is responsible for.
            counter.add(NULL, idx_end - idx_start)
            return

        # We partition the samples in two nodes on a given dimension,
        # with the middle point as a pivot.
        partition_node_indices(data_ptr, indices_ptr, idx_start, mid, idx_end, dim, n_features)

        self._left = consume Node(self._node_data_ptr, self._node_bounds_ptr, self._node_bounds_ptr_offset)
        self._right = consume Node(self._node_data_ptr, self._node_bounds_ptr, self._node_bounds_ptr_offset)

        # Recursing on both partitions.
        self._left.build_node(
            sync_method=NULL,
            node_index=2 * node_index + 1,
            data_ptr=data_ptr,
            indices_ptr=indices_ptr,
            leaf_size=leaf_size,
            n_features=n_features,
            dim=next_dim,
            idx_start=idx_start,
            idx_end=mid,
            counter=counter,
        )

        self._right.build_node(
            sync_method=NULL,
            node_index=2 * node_index + 2,
            data_ptr=data_ptr,
            indices_ptr=indices_ptr,
            leaf_size=leaf_size,
            n_features=n_features,
            dim=next_dim,
            idx_start=mid,
            idx_end=idx_end,
            counter=counter,
        )


cdef cypclass QueryActor activable:

    D_t * query_points
    D_t * data_ptr
    I_t * node_bounds_ptr
    NodeData_t * node_data_ptr
    I_t * indices_ptr
    NeighborsHeaps heaps
    I_t idx_worker
    I_t idx_start
    I_t idx_end
    I_t n_features
    I_t n_nodes
    I_t node_bounds_ptr_offset

    __init__(
        self,
        D_t * query_points,
        D_t * data_ptr,
        I_t * node_bounds_ptr,
        NodeData_t * node_data_ptr,
        I_t * indices_ptr,
        NeighborsHeaps heaps,
        I_t idx_worker,
        I_t idx_start,
        I_t idx_end,
        I_t n_features,
        I_t n_nodes,
        I_t node_bounds_ptr_offset,
    ):
        self.query_points = query_points
        self.data_ptr = data_ptr
        self.node_bounds_ptr = node_bounds_ptr
        self.node_data_ptr = node_data_ptr
        self.indices_ptr = indices_ptr
        self.heaps = heaps
        self.idx_worker = idx_worker
        self.idx_start = idx_start
        self.idx_end = idx_end
        self.n_features = n_features
        self.n_nodes = n_nodes
        self.node_bounds_ptr_offset = node_bounds_ptr_offset

    int query(self) except -1:
        cdef:
            I_t idx_pt
            D_t * query_point
            D_t reduced_dist_LB

        for idx_pt in range(idx_start, idx_end):
            query_point = self.query_points + idx_pt * self.n_features
            reduced_dist_LB = self.min_rdist(0, query_point)
            self._query_single_depthfirst(0, idx_pt, reduced_dist_LB)

    D_t min_rdist(self,
        I_t idx_node,
        D_t * pt,
    ) except -1:
        """Compute the minimum reduced-distance between a point and a node"""
        cdef:
            D_t d, d_lo, d_hi, node_min_j, node_max_j, rdist = 0.0
            I_t j

        for j in range(self.n_features):
            node_min_j = deref(self.node_bounds_ptr + idx_node + j * self.n_nodes)
            node_max_j = deref(self.node_bounds_ptr + idx_node + j * self.n_nodes + self.node_bounds_ptr_offset)

            d_lo = node_min_j - pt[j]
            d_hi = pt[j] - node_max_j
            # We use the following identity:
            #
            #       0.5 * (x + abs(x)) = 0.5 * max(x, 0)
            #
            # twice.
            d = 0.5 * ((d_lo + fabs(d_lo)) + (d_hi + fabs(d_hi)))
            rdist += d ** 2

        return rdist

    int _query_single_depthfirst(self,
        I_t idx_node,
        I_t idx_pt,
        D_t reduced_dist_LB,
    ) except -1:
        """Recursive Single-tree k-neighbors query, depth-first approach"""
        cdef NodeData_t node_info = self.node_data_ptr[idx_node]

        cdef D_t sq_dist, reduced_dist_LB_1, reduced_dist_LB_2
        cdef I_t i, idx_left_node, idx_right_node

        cdef D_t * query_point = query_points + idx_pt * self.n_features

        #------------------------------------------------------------
        # Case 1: query point is outside node radius:
        #         trim it from the query
        cdef D_t largest = heaps.largest(idx_pt)

        if reduced_dist_LB > largest:
            pass

        #------------------------------------------------------------
        # Case 2: this is a leaf node.  Update set of nearby points
        elif node_info.is_leaf:
            for i in range(node_info.idx_start, node_info.idx_end):
                sq_dist = sqeuclidean_dist(
                    x1=query_point,
                    x2=self.data_ptr + self.indices_ptr[i] * self.n_features,
                    k=self.n_features,
                )
                heaps.push(idx_pt, sq_dist, self.indices_ptr[i])

        #------------------------------------------------------------
        # Case 3: Node is not a leaf.  Recursively query subnodes
        #         starting with the closest
        else:
            idx_left_node = 2 * idx_node + 1
            idx_right_node = idx_left_node + 1
            reduced_dist_LB_1 = self.min_rdist(idx_left_node, query_point)
            reduced_dist_LB_2 = self.min_rdist(idx_right_node, query_point)

            # recursively query subnodes
            if reduced_dist_LB_1 <= reduced_dist_LB_2:
                self._query_single_depthfirst(idx_left_node, idx_pt, reduced_dist_LB_1)
                self._query_single_depthfirst(idx_right_node, idx_pt, reduced_dist_LB_2)
            else:
                self._query_single_depthfirst(idx_right_node, idx_pt, reduced_dist_LB_2)
                self._query_single_depthfirst(idx_left_node, idx_pt, reduced_dist_LB_1)
        return 0


cdef cypclass KDTree:
    """A KDTree based on asynchronous and parallel computations.

    It is similar to :class:`sklearn.neighbours.KDTree` but it is
    using Cython+ features for asynchronous computations.

    Asynchronuous and parallel computations can be used when
    constructing the nodes (see above) and when querying (WIP).

    This relies on a Cython+ runtime using actors.
    """

    I_t _n_samples
    I_t _n_features
    I_t _leaf_size
    I_t _n_levels
    I_t _n_nodes
    I_t _n_leafs

    active Node _root

    D_t *_data_ptr
    I_t *_indices_ptr
    NodeData_t *_node_data_ptr
    D_t * _node_bounds_ptr
    I_t _node_bounds_ptr_offset


    __init__(
        self,
        np.ndarray X,
        I_t leaf_size,
    ):
        cdef I_t i
        cdef I_t initialised = 0
        # Accessing _SC_NPROCESSORS_ONLN does not return the
        # effective number of threads which were assigned to
        # this task using `taskset(1)` for instance.
        #
        # This OpenMP API is a workable way to access it.
        cdef I_t num_workers = omp_get_max_threads()

        self._n_samples = X.shape[0]
        self._n_features = X.shape[1]
        self._leaf_size = leaf_size

        self._n_levels = <I_t> (log2(fmax(1, (self._n_samples - 1) / self._leaf_size)) + 1)
        self._n_nodes = <I_t> (2 ** (self._n_levels + 1))

        self._data_ptr = <D_t *> X.data
        self._indices_ptr = <I_t *> malloc(self._n_samples * sizeof(I_t))

        self._node_data_ptr = <NodeData_t *> malloc(self._n_nodes * sizeof(NodeData_t))
        self._node_bounds_ptr_offset = self._n_nodes * self._n_features

        # To be seen as a [2, n_nodes, d] with:
        #    - elements in [0, :, :] as min
        #    - elements in [1, :, :] as max
        self._node_bounds_ptr = <D_t *> malloc(2 * self._node_bounds_ptr_offset * sizeof(D_t))

        for i in range(self._n_samples):
            self._indices_ptr[i] = i

        # Recursively building the tree here
        global scheduler
        scheduler = Scheduler(num_workers)

        # This Counter is used as a way to implement a barrier for
        # the asynchronous construction of the tree.
        cdef active Counter counter = consume Counter()

        self._root = consume Node(self._node_data_ptr, self._node_bounds_ptr, self._node_bounds_ptr_offset)
        if self._root is NULL:
            printf("Error consuming node\n")

        # When object are activated (set as Actors), methods
        # are reified. When using those reified methods
        # a new argument is prepredend for a predicate,
        # which we aren't using here, hence the extra NULL.
        #
        # Also using this separate method allowing using actors
        # because __init__ can't be reified.
        self._root.build_node(
            sync_method=NULL,
            node_index=0,
            data_ptr=self._data_ptr,
            indices_ptr=self._indices_ptr,
            leaf_size=self._leaf_size,
            n_features=self._n_features,
            dim=0,
            idx_start=0,
            idx_end=self._n_samples,
            counter=counter,
        )

        # Waiting for the tree construction to end
        # Somewhat similar to a thread barrier
        while initialised < self._n_samples:
            initialised = counter.value(NULL).getIntResult()

        counter.reset(NULL)

    void __dealloc__(self):
        scheduler.finish()
        free(self._indices_ptr)
        free(self._node_data_ptr)
        free(self._node_bounds_ptr)

    void query(self,
        np.ndarray query_points,  # IN
        np.ndarray knn_indices,   # IN/OUT
        np.ndarray knn_distances, # IN/OUT
    ):
        cdef:
            I_t completed_queries = 0
            I_t i
            I_t n_query = query_points.shape[0]
            I_t n_features = query_points.shape[1]
            I_t n_neighbors = knn_indices.shape[1]
            D_t * _query_points = <D_t *> query_points.data
            D_t rdist_lower_bound
            I_t n_workers = omp_get_max_threads()
            I_t n_points_worker = <I_t> ceil(n_query / n_workers)
            I_t idx_start, idx_end, idx_worker
            NeighborsHeaps heaps = NeighborsHeaps(
                <I_t *> knn_indices.data,
                <D_t *> knn_distances.data,
                n_query,
                n_neighbors
            )
            # This Counter is used as a way to implement a barrier for
            # the asynchronous construction of the tree.
            active Counter counter = consume Counter()
            active QueryActor query_actor

        for idx_worker in range(n_workers):
            query_actor = consume QueryActor(
                query_points,
                self._node_bounds_ptr,
                self._node_data_ptr,
                self._indices_ptr,
                heaps,
                idx_worker,
                n_points_worker * idx_worker,
                min(n_points_worker * (idx_worker + 1), n_query),
                n_features,
                self._n_nodes,
                self._node_bounds_ptr_offset,
            )
            query_actor.query(NULL)

        heaps.sort()

cdef public int main() nogil:
    # Entry point for the compiled binary file
    printf("empty public int main() nogil:")
    return 0
